-----------------------------------------------------------
-- Test Drive Unlimited 2 CAR/RIM/BIKE/BUILDING mesh script
-- for 3D STUDIO MAX. Made with 3DSMAX 9, script should
-- work on max 5+ too. Rotate everything by +90deg on
-- X axis after importing Some materals and normals will not be
-- correct. Made by vagos21. e-mail: vagos21@yahoo.com
-- 
-- Run the script and select the extracted bnk folder For the
-- first run, you might also want to let it convert all
-- the textures too. Wait for a few secs, and you're done!
-----------------------------------------------------------

struct objesctstable(name, offset)
struct materialstable(name, offset)
struct instancestable(name, offset, matrix, materialsoffset, meshname)
struct GEOM(name, numprims, prims = #() )
struct PRIM(vboffset, iboffset, numverts, numinds, amesh)
struct hierarchytable(name, offset, parent, parentoffset, level, matrix)


matsarr = #() --materials array
geoms = #() --meshes array
--texpaths = #() --all textures included in sub-folders

macroScript TDU2_v2 category:"myscripts"
(
	undo off

	modelpath = getSavePath caption:"Select extracted TDU model folder"
	--commontexpath = getSavePath caption:"Select -Common_car- TDU folder"
	
	--print modelpath
	
	--convert all textures in folder to DDS
	convertTex = (queryBox "Do you want to convert all Textures in the folder?" title: "Texture Conversion" )
	if convertTex then(
		--get all textures included in folder and sub-folders
		for tex in (getFilesRecursive modelpath "*.2db") do
		(
			val = matchPattern tex pattern:"*cube*"
			if (not val) then --don't convert cubemaps
			(
				convertTexture tex
			)
		)
		print "textures converted"
	) 

	loadMeshes modelpath --load all meshes (3DG file)
	loadMaterials modelpath --load all materials (2DM file)
	buildScene modelpath--build scene from materials and meshes (load 3DD file)

	--finally rotate scene by 90degrees
	try
		(
		select $SCENE01
		rotate $ (eulerangles 90 0 0)
	)catch()

) --main script

fn buildScene filename =
(
	--open file
	path = getFiles (filename+"\\*.3dd")
	f = fopen path[1]  "rb"
	
	--jump MBNA
	fseek f 24 #seek_set
	MBNAsize = readlong f #unsigned
	fseek f (MBNAsize - 12) #seek_cur
	--jump OBJA
	fseek f 8 #seek_cur
	OBJAsize = readlong f #unsigned
	fseek f (OBJAsize - 12) #seek_cur
	--read INSA table
	fseek f 8 #seek_cur
	INSAsize = readlong f #unsigned
	fseek f 16 #seek_cur
	insanum = ((readlong f #unsigned)-16) / 16
	instances = #(insanum) --insa array
	for i=1 to insanum do
	(
		aname = ""
		for w=1 to 8 do (aname += (bit.intasChar(readByte f #unsigned)))
		aoffset = readlong f #unsigned
		fseek f 4 #seek_cur
		instances[i] = instancestable aname aoffset
	)
	
	--read corresponding GEOM and material names, also transformation matrix
	for i=1 to insanum do
	(	
		fseek f instances[i].offset #seek_set
		--read transformation matrix
		fseek f 16 #seek_cur
		transMatrix= matrix3 0
		transMatrix.row1=[readfloat f, readfloat f, readfloat f]
		fseek f 4 #seek_cur
		transMatrix.row2=[readfloat f, readfloat f, readfloat f]
		fseek f 4 #seek_cur
		transMatrix.row3=[readfloat f, readfloat f, readfloat f]
		fseek f 4 #seek_cur
		transMatrix.row4=[readfloat f, readfloat f, readfloat f]
		fseek f 4 #seek_cur
		--read material offset
		fseek f 4 #seek_cur
		amatoffset = readlong f #unsigned
		--read mesh name offset
		fseek f 8 #seek_cur
		meshoffset = readlong f #unsigned
		if (meshoffset == 0) then --it's a dummy
		(
			adummy = dummy name:(instances[i].name)
			adummy.boxsize = [0.1,0.1,0.1]
		)
		else
		(
			--read mesh name
			fseek f (meshoffset) #seek_set
			ameshname = ""
			for w=1 to 8 do (ameshname += (bit.intasChar(readByte f #unsigned)))
			--add info to instances table
			instances[i].meshname = ameshname
			instances[i].materialsoffset = amatoffset
	--print ameshname
		)
		instances[i].matrix = transMatrix
	)

	--clone meshes according to names and assign materials to primitives
	for i=1 to insanum do
	(
		for j=1 to geoms.count do
		(
			if (instances[i].meshname == geoms[j].name ) then -- we found an existing mesh
			(
				--first mesh is first primitive
				newmesh = copy geoms[j].prims[1].amesh
				--give name to new mesh 
				newmesh.Name = (instances[i].name)
				--goto materials list
				fseek f (instances[i].materialsoffset+16) #seek_set
				--find material
				amatname = ""
				for w=1 to 16 do (amatname += (bit.intasChar(readByte f #unsigned)))
				for mat=1 to matsarr.count do
				(
					if (amatname == matsarr[mat].name) then
					(
						--assign material to clone
						newmesh.material = matsarr[mat]
						exit
					)		
				)
				
				if geoms[j].numprims>1 then --attach the rest of the primitives
				(
					for k=2 to geoms[j].numprims do
					(
						amesh = copy geoms[j].prims[k].amesh
						--find material
						amatname = ""
						for w=1 to 16 do (amatname += (bit.intasChar(readByte f #unsigned)))
						for mat=1 to matsarr.count do
						(
							if (amatname == matsarr[mat].name) then
							(
								--assign material to clone
								amesh.material = matsarr[mat]
								exit
							)
						)
						meshop.attach newmesh amesh attachMat: #MatToID
					)
				)
				exit --go to next instance
			)
		)
	)
	
	
	--delete ALL previous meshes used!!!
	for i=1 to geoms.count do
	(
		for j=1 to geoms[i].numprims do
		(
			delete geoms[i].prims[j].amesh
		)
	)
	--clean up arrays
	for i=1 to geoms.count do (deleteItem geoms 1)
	for i=1 to matsarr.count do (deleteItem matsarr 1)
	
	
	hierarchy = #() --hierarchy table
	--build scene hierarchy
	fseek f (MBNAsize+OBJAsize+INSAsize+16) #seek_set
	type = "" --read if HIEA exists
	for w=1 to 4 do (type += (bit.intasChar(readByte f #unsigned)))
	if (type == "HIEA") then --scene has hierarchy
	(
		fseek f 28 #seek_cur --go to hierarchy table
		--read hierarchy table and build structure
		for i=1 to insanum do
		(
			aname = ""
			for w=1 to 8 do (aname += (bit.intasChar(readByte f #unsigned)))
			aoffset = readlong f #unsigned
			fseek f 4 #seek_cur
			ahiea = hierarchytable aname aoffset
			append hierarchy ahiea
		)
		for i=1 to insanum do --find parent name
		(
			fseek f (hierarchy[i].offset+28) #seek_set
			parentoffset=readlong f #unsigned
			hierarchy[i].parentoffset = parentoffset
			hierarchy[i].matrix = instances[i].matrix
			for j=1 to insanum do
			(
				if (hierarchy[j].offset==parentoffset) then (hierarchy[i].parent=hierarchy[j].name; exit)
			)		
		)
		--sort hierarchy array by parent offset (stonesort :P)
		for i = 1 to insanum-1 do
		(
			for j = i+1 to insanum do
    		(
      			if (hierarchy[i].parentoffset > hierarchy[j].parentoffset) then
      			(
        			temphierarchy = hierarchy[j];
        			hierarchy[j] = hierarchy[i];
        			hierarchy[i] = temphierarchy;
      			)
    		)
		)
		--find level for each node
		hierarchy[1].level=1
		for i=2 to insanum do
		(
			for j=1 to i do
			(
				if (hierarchy[i].parent == hierarchy[j].name) then ( hierarchy[i].level = hierarchy[j].level+1 ;exit)
			)
		)
		--find max number of levels
		maxlevel=1
		for i=1 to insanum do
		(
			if (hierarchy[i].level > maxlevel) then (maxlevel = hierarchy[i].level)
		)
		--apply hierarchy structure to loaded nodes
		for lev=maxlevel to 2 by -1 do
		(
			--transform every object in that level
			for i=1 to insanum do
			(
				if (hierarchy[i].level==lev) then
				(
					(getNodeByName hierarchy[i].name).transform = hierarchy[i].matrix
				)	
			)
			--add current level object to parent
			for i=1 to insanum do
			(
				if (hierarchy[i].level==lev) then
				(
					attachObjects (getNodeByName hierarchy[i].parent) (getNodeByName hierarchy[i].name) move:false
				)
			)
		)
		
	) --hierarchy if

	
	--close file
	fclose f
	
	print "Scene created!"
) --buildScene

fn loadMaterials filename =
(
	--open file
	path = getFiles (filename+"\\*.2dm")
	f = fopen path[1]  "rb"
	
	--read materials table
	fseek f 44 #seek_set
	materialsnum = ((readlong f #unsigned)-16) / 16
	print (materialsnum as string+" materials found.")
	materials = #(materialsnum)
	for i=1 to materialsnum do
	(
		aname = ""
		for w=1 to 8 do (aname += (bit.intasChar(readByte f #unsigned)))
		aoffset = readlong f #unsigned
		fseek f 4 #seek_cur
		materials[i] = materialstable aname aoffset
	--print materials[i]
	)
	
	--apo to MAT
	for i=1 to materialsnum do
	(
		fseek f materials[i].offset #seek_set
		
		--read number of textures used
		fseek f 268 #seek_cur
		numproperties = readshort f #unsigned
		fseek f 2 #seek_cur
		
		--print ("Loading Material: "+materials[i].name)
		amaterial = standard showInViewport:true Name: materials[i].name
		
		for j=1 to numproperties do
		(
			paramType = ""
			for w=1 to 8 do (paramType += (bit.intasChar(readByte f #unsigned)))
			paramValue = ""
			for w=1 to 8 do (paramValue += (bit.intasChar(readByte f #unsigned)))
			fseek f 16 #seek_cur
			

			if (paramValue.count > 0) then
			(
				case substring paramType 3 (paramType.count-2) of
				(
					"TAIL": --diffuse (UV1)
						(	
							adiffusemap = Bitmaptexture fileName: (filename+"/maps/"+paramValue+".dds")
							amaterial.diffuseMap=(adiffusemap)
							--print ("    DIFFUSE MAP: "+adiffusemap.filename)
						)
					"RMAL": --normal
						(
							anormalmap = Bitmaptexture fileName: (filename+"/maps/"+paramValue+".dds")
							amaterial.bumpMap=(anormalmap)
							--print ("    NORMAL MAP: "+anormalmap.filename)
						)				
					"OSS": --gloss
						(
							aglossmap = Bitmaptexture fileName: (filename+"/maps/"+paramValue+".dds")
							amaterial.specularLevelMap =(aglossmap )
							--print ("    GLOSS MAP: "+aglossmap.filename)
						)
					"FLECTI": --reflection
						()
					"XLIGHT": --tex light
						()
					"OSSLIG": --gloss light
						()
					"TAIL2": --texture2 (UV2)
						()
					"ADOW": --shadow
						()
					"LOR": --color
						()
				)
			)
		)
		
		amaterial.twoSided = on	
		amaterial.ambient = color 0 0 0
		amaterial.Diffuse = color 128 128 128
		amaterial.Specular = color 255 255 255
		--append loaded material to materials array
		append matsarr amaterial
	)

	print "Materials Loaded!"
	--close materials file
	fclose f
) --loadMaterials

fn loadMeshes filename =
(
	--open file
	path = getFiles (filename+"\\*.3dg")
	f = fopen path[1]  "rb"
	
	--read number of meshes	
	fseek f 44 #seek_set
	meshesnum = ((readlong f #unsigned)-16) / 16

	--read objects table
	objects = #(meshesnum)
	for i=1 to meshesnum do
	(
		aname = ""
		for w=1 to 8 do (aname += (bit.intasChar(readByte f #unsigned)))
		aoffset = readlong f #unsigned
		fseek f 4 #seek_cur
		objects[i] = objesctstable aname aoffset
	)

	--read GEOMS
	for i=1 to meshesnum do
	(
		primarr = #()
		startoffset = objects[i].offset
		fseek f (startoffset + 8) #seek_set
		geomsize = (readlong f #unsigned)
		fseek f 4 #seek_cur
		numprims=0
		do 
		(
			--read PRIM or BONE or anything else...
			primname = ""
			for w=1 to 4 do (primname += (bit.intasChar(readByte f #unsigned)))
			--read PRIM or BONE size
			fseek f 8 #seek_cur
			structsize =  readlong f #unsigned
			--if it's a PRIM, read it
			if (primname == "PRIM") then
			(
				numprims = numprims +1
				iboffset = readlong f #unsigned
				vboffset = readlong f #unsigned
				fseek f 20 #seek_cur
				numverts = readlong f #unsigned
				fseek f 8 #seek_cur
				numinds = readlong f #unsigned
				fseek f 20 #seek_cur --jump remaining bytes
				aprim = PRIM vboffset iboffset numverts numinds
				append primarr aprim
			)else(
				fseek f (structsize-16) #seek_cur
			)
		) while ((ftell f)-startoffset < geomsize)
		ageom = GEOM objects[i].name numprims primarr
--print ageom
		append geoms ageom
	)

	--read primitives as meshes and attach them together in a GEOM
	for i=1 to meshesnum do
	(
		for j=1 to geoms[i].numprims do
		(
			farr = #() --faces array
			varr = #() --vertex positions array
			UVWarr = #() --vertex UV coords array
			
			--read vertex buffer
			readVB f geoms[i].prims[j].vboffset geoms[i].prims[j].numverts varr UVWarr
			
			--read index buffer
			readIB f geoms[i].prims[j].iboffset geoms[i].prims[j].numinds farr
			
			--build mesh	
			geoms[i].prims[j].amesh = mesh vertices:varr faces:farr name:geoms[i].name
			
			--apply uvw tex coords to mesh
			applyuvw geoms[i].prims[j].amesh UVWarr
		)
	)
	
	--close 3DG file
	fclose f
	
	print "Meshes loaded!"
	
) --loadMeshes

fn convertTexture filename =
(
	--DDS HEADER info from http://msdn.microsoft.com/archive/default.asp?url=/archive/en-us/directx9_c/directx/graphics/reference/ddsfilereference/ddsfileformat.asp
	--DDS HEADER values from http://wiki.answers.com/Q/How_do_you_write_a_header_for_a_128x128x128_dds_volume_texture_file_with_no_mipmaps_and_does_it_have_to_be_written_in_binary
	
	--DDS header stuff
	DDSD_CAPS =			0x00000001
	DDSD_HEIGHT =		0x00000002
	DDSD_WIDTH =		0x00000004
	DDSD_PITCH =		0x00000008
	DDSD_PIXELFORMAT =	0x00001000
	DDSD_MIPMAPCOUNT =	0x00020000
	DDSD_MIPMAPCOUNT =	0x00020000
	DDSD_LINEARSIZE =	0x00080000
	DDSD_DEPTH =		0x00800000
	
	DDPF_ALPHAPIXELS =	0x00000001
	DDPF_FOURCC =		0x00000004
	DDPF_RGB =			0x00000040
	
	DDSCAPS_COMPLEX= 	0x00000008
	DDSCAPS_TEXTURE =	0x00001000
	DDSCAPS_MIPMAP =	0x00400000
	
	DDSCAPS2_CUBEMAP =			0x00000200
	DDSCAPS2_CUBEMAP_POSITIVEX=	0x00000400
	DDSCAPS2_CUBEMAP_NEGATIVEX=	0x00000800
	DDSCAPS2_CUBEMAP_POSITIVEY=	0x00001000
	DDSCAPS2_CUBEMAP_NEGATIVEY=	0x00002000
	DDSCAPS2_CUBEMAP_POSITIVEZ=	0x00004000
	DDSCAPS2_CUBEMAP_NEGATIVEZ=	0x00008000
	DDSCAPS2_VOLUME =			0x00200000
	
	DXT1 = 0x31545844
	DXT2 = 0x32545844
	DXT3 = 0x33545844
	DXT4 = 0x34545844
	DXT5 = 0x35545844
	
	ddsd = #(18) --dds description
	ddpf = #(8) --dds pixel format
	ddsCaps = #(4) --dds caps
	
	--read texture header stuff
	fin = fopen filename "rb"
	--read texture size
	fseek fin 28 #seek_set
	texsize = (readlong fin #unsigned)-64
	--read texture name
	texname = ""
	for w=1 to 8 do (texname += (bit.intasChar(readByte fin #unsigned)))
	--read texture width and height
	width = readshort fin
	height = readshort fin
	fseek fin 2 #seek_cur
	--read number of mipmaps
	mipmapcount = readbyte fin #unsigned
	fseek fin 1 #seek_cur
	--read type of texture (DXT, uncompressed, cubemap etc...
	textype = readlong fin #unsigned
	fseek fin 28 #seek_cur
	--load texture data into array
	texdata = #(texsize)
	for i = 1 to texsize do texdata[i]= (readbyte fin #unsigned)
	--close source file
	fclose fin
	
	print ("Converting texture: "+filename+", Size: "+width as string+"x"+height as string+" mipmaps: "+mipmapcount as string)
	
	--build dds header from data
	dwMagic = 0x20534444 --"DDS"
	ddsd[1] = 124
	ddsd[3] = height
	ddsd[4] = width
	ddsd[6] = 32
	ddsd[7] = mipmapcount
	for i = 8 to 18 do ddsd[i]=0
	ddpf[1] = 32
	ddpf[4] = 32
	ddpf[5] = 0x00ff0000
	ddpf[6] = 0x0000ff00
	ddpf[7] = 0x000000ff
	ddpf[8] = 0xff000000
	ddsCaps[1] = DDSCAPS_MIPMAP
	ddsCaps[2] = 0
	ddsCaps[3] = 0
	ddsCaps[4] = 0
	--case specific header stuff
	case textype of
	(
		132:( --working
			--print "Texture format is DXT1 " --DXT1
			ddsd[2] = bit.or (bit.or (bit.or(bit.or (bit.or DDSD_CAPS DDSD_HEIGHT) DDSD_WIDTH) DDSD_PIXELFORMAT) DDSD_MIPMAPCOUNT) DDSD_LINEARSIZE
			ddpf[3] = DXT1
			ddsd[5] = texsize
			ddpf[2] = DDPF_FOURCC
			)
		136:( --working
			--print "Texture format is DXT5" --DXT5
			ddsd[2] = bit.or (bit.or (bit.or(bit.or (bit.or DDSD_CAPS DDSD_HEIGHT) DDSD_WIDTH) DDSD_PIXELFORMAT) DDSD_MIPMAPCOUNT) DDSD_LINEARSIZE
			ddpf[3] = DXT5
			ddsd[5] = texsize
			ddpf[2] = DDPF_FOURCC
			)
		144:(
			--print "Texture format is Uncompressed" --uncompressed RGBA!
			ddsd[2] = bit.or(bit.or(bit.or(bit.or (bit.or DDSD_CAPS DDSD_HEIGHT) DDSD_WIDTH) DDSD_PIXELFORMAT) DDSD_MIPMAPCOUNT) DDSD_PITCH
			ddpf[3] = 0
			ddsd[5] = width*4
			ddpf[2] = DDPF_RGB
			)
	)
	
	--open output file
	fout = fopen ((substring filename 1 (filename.count-4)) +".dds") "wb"
	--write header to file
	writelong fout dwMagic
	for i = 1 to 18 do (writeLong fout ddsd[i])
	for i = 1 to 8  do (writeLong fout ddpf[i])
	for i = 1 to 4  do (writeLong fout ddsCaps[i])
	writeLong fout 0
	--write texture data to file
	for i = 1 to texsize do (writeByte fout texdata[i] #unsigned)
	--flush output file
	fflush fout
	--close output file
	fclose fout

) --convertTextures

fn applyuvw anode auvarray	= --applies UVW to node
(
	if (auvarray.count > 0) then (
		anode.numtverts = anode.numverts
		buildtvfaces anode
		for vi = 1 to anode.numverts do (
			settvert anode vi auvarray[vi]
		)				
		allfaces = #()
		for fi = 1 to anode.numfaces do (
			crtface = getface anode fi
			append allfaces crtface
		)
		for fi = 1 to anode.numfaces do (
			settvface anode fi allfaces[fi]
		)
	)
) -- apply UV

fn readVB f offset numverts varr UVWarr =
(
	FVF = #() --Flexible Vertex Format header array
	
	--go to vertex buffer
	fseek f offset #seek_set
	--jump a few bytes
	fseek f 20 #seek_cur
	--read FVF
	for inp = 1 to 28 do (
		FVF[inp] = readbyte f
	)
	hasNormal  = FVF[3]
	hasColor = FVF[5]
	numUVs = FVF[9]
	hasFirstQQ = FVF[11]
	hasSecondQQ = FVF[13]
	hasUVinPOS = FVF[26]
		
	--read positions and UVs
	if (hasUVinPOS == 16) then
	(
		--UVs are with the positions
		for v = 1 to numverts do
		(
			--read positions
			avert = [0.,0.,0.]
			avert.x = (readfloat f)
			avert.y = (readfloat f)
			avert.z = (readfloat f)
			append varr avert
			--read uv sets
			if (numUVs > 0) then
			(
				aUVW = [readfloat f, -readfloat f, 0.0]
				append UVWarr aUVW
				--jump the rest of uvsets
				if (numUVs > 1) then fseek f (8*(numUVs-1)) #seek_cur
			)
		)
	)else(
		--positions are alone in first set
		for v = 1 to numverts do
		(
			--read positions
			avert = [0.,0.,0.]
			avert.x = (readfloat f)
			avert.y = (readfloat f)
			avert.z = (readfloat f)
			append varr avert
		)
	)
	
	---read normals and colors and UVs
	for v = 1 to numverts do
	(
		--jump normal if it exists
		if (hasNormal==1) then (fseek f 8 #seek_cur)
		
		--jump color if it exists
		if (hasColor==1) then (fseek f 4 #seek_cur)
		
		--read uv
		if (hasUVinPOS == 1) then
		(
			if (numUVs > 0) then
			(
				aUVW = [readfloat f, -readfloat f, 0.0]
				append UVWarr aUVW
				--jump the rest of uvsets
				if (numUVs > 1) then fseek f (8*(numUVs-1)) #seek_cur
			)
		)
		
		--jump first QQ if it exists
		if (hasFirstQQ==1) then (fseek f 8 #seek_cur)
		
		--jump first QQ if it exists
		if (hasSecondQQ==1) then (fseek f 8 #seek_cur)
		
	)
) --readVB

fn readIB f offset numinds farr =
(
	iarr = #() --indices array
	
	--read index buffer
	fseek f offset #seek_set
	fseek f 32 #seek_cur
	for v = 1 to numinds do
	(
		aind = readshort f #unsigned
		append iarr (aind+1)
	)
			
	--build faces array from indices (convert strip to list)
	for ind = 1 to numinds-2 do
	(
		fa = iarr[ind]
		fb = iarr[ind+1]
		fc = iarr[ind+2]		
		if ((fa!=fb)and(fa!=fc)and(fc!=fb)) then
		(
			if ( mod ind 2 == 0 ) then
			(
				aface = [fb,fa,fc]
			)else	(
				aface = [fa,fb,fc]								
			)
			append farr aface
		)
	)
) --readIB

fn getFilesRecursive root pattern =
(
	dir_array = GetDirectories (root+"/*")
	for d in dir_array do
	join dir_array (GetDirectories (d+"*"))
	my_files = #()
	for f in dir_array do
	join my_files (getFiles (f + pattern))
	my_files 
)
